/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/slab.h>

#include "cow_range.h"

#define PTRS_PER_RANGE 32  /* must be power of 2 */

struct cow_range {
	struct rb_node node;
	u32 start;
	struct cow_buffer *cows[PTRS_PER_RANGE];
};

struct cow_buffer *cow_search_range(struct rb_root *root, u32 index)
{
	struct rb_node *node = root->rb_node;

	unsigned int offset = index & (PTRS_PER_RANGE-1);
	u32 start = index - offset;

	while (node) {
		struct cow_range *range = container_of(node, struct cow_range, node);

		if (start < range->start)
			node = node->rb_left;
		else if (start >= range->start + PTRS_PER_RANGE)
			node = node->rb_right;
		else
			return range->cows[offset];
	}
	return NULL;
}

int cow_add_range(struct rb_root *root, u32 index, struct cow_buffer *cow)
{
	struct rb_node **new = &(root->rb_node), *parent = NULL;
	struct cow_range *range;

	unsigned int offset = index & (PTRS_PER_RANGE-1);
	u32 start = index - offset;

	while (*new) {
		range = container_of(*new, struct cow_range, node);

		parent = *new;
		if (index < range->start) {
			new = &((*new)->rb_left);
		} else if (index >= range->start + PTRS_PER_RANGE) {
			new = &((*new)->rb_right);
		} else {
			range->cows[offset] = cow;
			return 0;
		}
	}

	/* Add new node and rebalance tree. */
	range = kzalloc(sizeof(struct cow_range), GFP_KERNEL);
	if (!range)
		return -ENOMEM;

	RB_CLEAR_NODE(&range->node);
	range->start = start;
	rb_link_node(&range->node, parent, new);
	rb_insert_color(&range->node, root);

	return 0;
}
