/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#undef TRACE_SYSTEM
#define TRACE_SYSTEM stcvsm
#if !defined(STCVSM_DEBUG_H_TRACE) || defined(TRACE_HEADER_MULTI_READ)
#define STCVSM_DEBUG_H_TRACE

#ifdef STCVSM_TRACEPOINTS
#include <linux/tracepoint.h>

DECLARE_EVENT_CLASS(reqfn_io,
	TP_PROTO(int major, int minor, unsigned long long sector, unsigned int sectors),
	TP_ARGS(major, minor, sector, sectors),
	TP_STRUCT__entry(
		__field(int, major)
		__field(int, minor)
		__field(unsigned long long, sector)
		__field(unsigned int, sectors)
	),
	TP_fast_assign(
		__entry->major = major;
		__entry->minor = minor;
		__entry->sector = sector;
		__entry->sectors = sectors;
	),
	TP_printk("[%d:%d][%llu] sectors %u\n",
		__entry->major, __entry->minor,
		__entry->sector, __entry->sectors)
);

DECLARE_EVENT_CLASS(cow_state_trans,
	TP_PROTO(void *cow, void *bio, int old_state, int new_state, unsigned long in_state),
	TP_ARGS(cow, bio, old_state, new_state, in_state),
	TP_STRUCT__entry(
		__field(void *, cow)
		__field(void *, bio)
		__field(int, old_state)
		__field(int, new_state)
		__field(unsigned long, in_state)
	),
	TP_fast_assign(
		__entry->cow = cow;
		__entry->bio = bio;
		__entry->old_state = old_state;
		__entry->new_state = new_state;
		__entry->in_state = in_state;
	),
	TP_printk("%p(%p) %d->%d (%lu)",
		__entry->cow, __entry->bio,
		__entry->old_state, __entry->new_state,
		__entry->in_state)
);

#define STC_DEFINE_EVENT DEFINE_EVENT

#else
// No-op the trace functions
#define TP_PROTO(args...)	args
#define TP_ARGS(args...)	args
#define TP_CONDITION(args...)	args
#define STC_DEFINE_EVENT(tmpl, name, proto, args) \
	static inline void trace_ ## name(proto) { }

#endif

#define DEFINE_EVENT_REQFN_IO(name) \
	STC_DEFINE_EVENT(reqfn_io, name, \
		TP_PROTO(int major, int minor, unsigned long long sector, unsigned int sectors), \
		TP_ARGS(major, minor, sector, sectors) \
	);

#define DEFINE_EVENT_COW_STATE_TRANS(name) \
	STC_DEFINE_EVENT(cow_state_trans, name, \
		TP_PROTO(void *cow, void *bio, int old_state, int new_state, unsigned long in_state), \
		TP_ARGS(cow, bio, old_state, new_state, in_state) \
	);

DEFINE_EVENT_REQFN_IO(snap_ctl_reqfn_io);
DEFINE_EVENT_REQFN_IO(snap_read_reqfn_io);
DEFINE_EVENT_REQFN_IO(snap_write_reqfn_io);
DEFINE_EVENT_REQFN_IO(orig_ctl_reqfn_io);
DEFINE_EVENT_REQFN_IO(orig_read_reqfn_io);
DEFINE_EVENT_REQFN_IO(orig_write_reqfn_io);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_new);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_filling);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_pagedin);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_pagedout);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_pagingin);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_error);
DEFINE_EVENT_COW_STATE_TRANS(cow_state_trans_barrier);

#endif // STCVSM_DEBUG_H_TRACE

#ifdef STCVSM_TRACEPOINTS
#undef TRACE_INCLUDE_PATH
#define TRACE_INCLUDE_PATH .
#define TRACE_INCLUDE_FILE debug
#include <trace/define_trace.h>
#endif

#ifndef STCVSM_DEBUG_H
#define STCVSM_DEBUG_H

#include <linux/sched.h>
#include <linux/version.h>
#ifdef HAVE_KERN_LEVELS_H
#include <linux/kern_levels.h>
#endif

#ifndef LOGLEVEL_ERR
#define LOGLEVEL_ERR 3
#endif
#ifndef LOGLEVEL_WARNING
#define LOGLEVEL_WARNING 4
#endif
#ifndef LOGLEVEL_NOTICE
#define LOGLEVEL_NOTICE 5
#endif
#ifndef LOGLEVEL_INFO
#define LOGLEVEL_INFO 6
#endif
#ifndef LOGLEVEL_DEBUG
#define LOGLEVEL_DEBUG 7
#endif
#ifndef LOGLEVEL_TRACE
#define LOGLEVEL_TRACE 8
#endif

#ifndef STCVSM_LOG_LEVEL
#ifdef DEBUG
#define STCVSM_LOG_LEVEL LOGLEVEL_DEBUG
#else
#define STCVSM_LOG_LEVEL LOGLEVEL_NOTICE
#endif
#endif

#define LOG_COWS 0
#define LOG_BLKRW 1
#define LOG_MODULE 2
#define LOG_DEBUG 3
#define LOG_FS 4
#define LOG_PROC 5
#define LOG_SNAPSHOT 6
#define LOG_VDIFF 7
#define LOG_REFS 8
#define LOG_CTL 9
#define LOG_FILTER 10
#define LOG_IO 11
#define LOG_SETS 12
#define LOG_GUID 12
#define LOG_MAX 13

void stcvsm_logger(int name, int level, const char *format, ...)
	__attribute__((format(printf, 3, 4)));

/* Log the task_struct and CPU#?  Helps make sense of concurrency bugs. */
#ifdef LOG_CURRENT_TASK
#define LOG_CURTASK (current ? current->pid : -1),
#ifdef HAVE_TASK_ON_CPU
#define LOG_CURFMT "%5d(%d) "
#ifdef CONFIG_SMP
#define LOG_CURCPU (current ? current->on_cpu : -1),
#else
#define LOG_CURCPU 0,
#endif
#else /* !HAVE_TASK_ON_CPU */
#define LOG_CURFMT "%5d "
#define LOG_CURCPU
#endif
#else /* ! LOG_CURRENT_TASK */
#define LOG_CURTASK
#define LOG_CURCPU
#define LOG_CURFMT
#endif

/* Log function name? */
#ifdef LOG_FUNC_NAME
#define LOG_FUNCFMT "%s: "
#define LOG_FUNCSTR , __func__
#else
#define LOG_FUNCFMT
#define LOG_FUNCSTR
#endif

#define LOG_FORMAT(TYPE, NAME, fmt) TYPE "stcvsm " LOG_CURFMT "%-8s " LOG_FUNCFMT fmt, \
		LOG_CURTASK LOG_CURCPU NAME+4 LOG_FUNCSTR

#if STCVSM_LOG_LEVEL >= LOGLEVEL_TRACE
#define stcvsm_trace(NAME, fmt, ...) \
		stcvsm_logger(NAME, LOGLEVEL_TRACE, LOG_FORMAT(KERN_DEBUG, #NAME, fmt), ##__VA_ARGS__)
#else
#define stcvsm_trace(NAME, fmt, ...)
#endif
#if STCVSM_LOG_LEVEL >= LOGLEVEL_DEBUG
#define stcvsm_debug(NAME, fmt, ...) \
		stcvsm_logger(NAME, LOGLEVEL_DEBUG, LOG_FORMAT(KERN_DEBUG, #NAME, fmt), ##__VA_ARGS__)
#else
#define stcvsm_debug(NAME, fmt, ...)
#endif
#if STCVSM_LOG_LEVEL >= LOGLEVEL_INFO
#define stcvsm_info(NAME, fmt, ...) \
		stcvsm_logger(NAME, LOGLEVEL_INFO, LOG_FORMAT(KERN_INFO, #NAME, fmt), ##__VA_ARGS__)
#else
#define stcvsm_info(NAME, fmt, ...)
#endif
#if STCVSM_LOG_LEVEL >= LOGLEVEL_NOTICE
#define stcvsm_notice(NAME, fmt, ...) \
		stcvsm_logger(NAME, LOGLEVEL_NOTICE, LOG_FORMAT(KERN_NOTICE, #NAME, fmt), ##__VA_ARGS__)
#else
#define stcvsm_notice(NAME, fmt, ...)
#endif
#if STCVSM_LOG_LEVEL >= LOGLEVEL_WARNING
#define stcvsm_warning(NAME, fmt, ...) \
		stcvsm_logger(NAME, LOGLEVEL_WARNING, LOG_FORMAT(KERN_WARNING, #NAME, fmt), ##__VA_ARGS__)
#else
#define stcvsm_warning(NAME, fmt, ...)
#endif
#if STCVSM_LOG_LEVEL >= LOGLEVEL_ERR
#define stcvsm_err(NAME, fmt, ...) \
		stcvsm_logger(NAME, LOGLEVEL_ERR, LOG_FORMAT(KERN_ERR, #NAME, fmt), ##__VA_ARGS__)
#else
#define stcvsm_err(NAME, fmt, ...)
#endif

#ifdef RUNTIME_LOG
void set_log_level(int type, int level);
#endif

#define printk_uuid(level, note, uuid) \
	stcvsm_##level(LOG_GUID, \
			"%s: {%08X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X}\n", \
			note, \
			*(u32 *)&(uuid)[0], \
			(uuid)[5], \
			(uuid)[4], \
			(uuid)[7], \
			(uuid)[6], \
			(uuid)[8], \
			(uuid)[9], \
			(uuid)[10], \
			(uuid)[11], \
			(uuid)[12], \
			(uuid)[13], \
			(uuid)[14], \
			(uuid)[15])


#define stcvsm_atomic_inc(v) \
	({ \
		atomic_inc(v);\
		stcvsm_info(LOG_REFS, "atomic_inc: " #v " %d\n", atomic_read(v)); \
	})
#define stcvsm_atomic_dec(v) \
	({ \
		atomic_dec(v);\
		stcvsm_info(LOG_REFS, "atomic_dec: " #v " %d\n", atomic_read(v)); \
	})
#define stcvsm_atomic_dec_and_lock(v, l) \
	({ \
		int r = atomic_dec_and_lock(v, l); \
		stcvsm_info(LOG_REFS, "atomic_dec_and_lock: " #v " " #l " %d\n", r); \
		r; \
	})
#define stcvsm_atomic_dec_and_test(v) \
	({ \
		int r = atomic_dec_and_test(v); \
		stcvsm_info(LOG_REFS, "atomic_dec_and_test: " #v " %d\n", r); \
		r; \
	})

/* Types of errors that can be injected */
#define ERRINJ_ALLOC            0
#define ERRINJ_BLKDEV_FREEZE    1
#define ERRINJ_BDGET            2
#define ERRINJ_BIO_READ         3
#define ERRINJ_BIO_WRITE        4
#define ERRINJ_COW_IO           5
#define ERRINJ_SNAPSHOT_ALREADY 6
#define ERRINJ_SNAPSHOT_MINOR   7
#define ERRINJ_NUM_TYPES        8

#ifdef ERRINJ
int stcvsm_error_inject(unsigned int type, const char *file, unsigned int line);
#define ERROR_INJECT(type) stcvsm_error_inject(type, __FILE__, __LINE__)
#else
#define ERROR_INJECT(type) 0
#endif

#endif
