#!/bin/bash

KERNEL=$1
KDIR=/lib/modules/$KERNEL/build
SYSTEM_MAP=/boot/System.map-$KERNEL

sudo=
if [ $(id -u) -gt 0 ]; then
	sudo=$(which sudo)
fi

if [ ! -e $KDIR ]; then
	echo "Folder '$KDIR' does not exist" >&2
	exit 1
fi

# 2.6
# make_request_fn has changed types a few times, and is then removed in 5.9
grep -q "typedef int .make_request_fn" $KDIR/include/linux/blkdev.h && echo "#define HAVE_INT_MAKE_REQUEST"

# blk_alloc_queue starts with parameters (gfp_t
grep -qw 'blk_alloc_queue.gfp_t' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BLK_ALLOC_GFP_T"
# 3.2
grep -q "typedef void .make_request_fn" $KDIR/include/linux/blkdev.h && echo "#define HAVE_VOID_MAKE_REQUEST"
# 3.4
grep -qw kmalloc_array $KDIR/include/linux/slab.h && echo "#define HAVE_KMALLOC_ARRAY"
# 3.6
grep -qw part_nr_sects_read $KDIR/include/linux/genhd.h && echo "#define HAVE_PART_NR_SECTS_READ"
test -e $KDIR/include/linux/kern_levels.h && echo "#define HAVE_KERN_LEVELS_H"
# 3.7
grep -qw bio_clone_bioset $KDIR/include/linux/bio.h && echo "#define HAVE_BIO_CLONE_BIOSET"
# 3.8
grep -qw prandom_u32 $KDIR/include/linux/random.h && echo "#define HAVE_PRANDOM"
# 3.10
grep -qw PDE_DATA $KDIR/include/linux/proc_fs.h && echo "#define HAVE_PDE_DATA"
# 3.12
grep -qw 'struct blk_mq_ops' $KDIR/include/linux/blk-mq.h && echo "#define HAVE_BLK_MQ_OPS"
# 3.14
grep -qw bio_clone_fast $KDIR/include/linux/bio.h && echo "#define HAVE_BIO_CLONE_FAST"
grep -qsw bvec_iter $KDIR/include/linux/blk_types.h && echo "#define HAVE_IMMUTABLE_BIOVEC"
# 4.3
grep -qs "int[[:space:]]*bi_error;" $KDIR/include/linux/blk_types.h && echo "#define HAVE_BIO_BI_ERROR"
# 4.4
grep -q "submit_bio.int" $KDIR/include/linux/fs.h && echo "#define HAVE_SUBMIT_BIO_RW"
grep -q "typedef blk_qc_t .make_request_fn" $KDIR/include/linux/blkdev.h && echo "#define HAVE_BLK_QC_T_MAKE_REQUEST"
# 4.8
grep -qs "int[[:space:]]*bi_opf;" $KDIR/include/linux/blk_types.h && echo "#define HAVE_BIO_BI_OPF"
# 4.11
grep -q "int[[:space:]]*on_cpu;" $KDIR/include/linux/sched.h && echo "#define HAVE_TASK_ON_CPU"
# 4.13
grep -qs "typedef.*blk_status_t;" $KDIR/include/linux/blk_types.h && echo "#define HAVE_BLK_STATUS_T"
grep -qw "struct wait_queue_entry" $KDIR/include/linux/wait.h && echo "#define HAVE_WAITQUEUE_ENTRY"
grep -qws "bio_set_dev" $KDIR/include/linux/bio.h && echo "#define HAVE_BIO_SET_DEV"
# 4.14
# Removed bi_bdisk in favor of bi_disk, and in 5.12, bi_disk was removed in favor of bi_bdisk
grep -qs "struct gendisk.**bi_disk;" $KDIR/include/linux/blk_types.h && echo "#define HAVE_BIO_GENDISK"
# 4.18
# Since 2.6, fs_bio_set was a pointer. In 4.18, it dropped the pointer.
grep -qw '^extern struct bio_set fs_bio_set' $KDIR/include/linux/bio.h && echo "#define HAVE_FS_BIO_SET_NO_PTR"
# get_budget_fn added, but then removed in 5.9, and declared as taking just a request_queue
# RedHat's backport keeps both implementations decided by __GENKSYMS__
grep -qw 'get_budget_fn..struct blk_mq_hw_ctx' $KDIR/include/linux/blk-mq.h && echo "#define HAVE_GET_BUDGET_FN_CTX"
grep -qw 'get_budget_fn..struct request_queue' $KDIR/include/linux/blk-mq.h && echo "#define HAVE_GET_BUDGET_FN_Q"
# 4.19
grep -qw 'struct blkcg_gq.*\*bi_blkg' $KDIR/include/linux/blk_types.h && echo "#define HAVE_BIO_BI_BLKG"
# 5.0
# Since 2.6, in_flight was part of hd_struct. Since 5.0 it's part of disk_stats
# TODO: We can probably restore this feature by finding the disk_status for our hd_struct.
# TODO: Perhaps we can start calling generic_start_io_acct?
grep -qw 'atomic_t in_flight\[2\];' $KDIR/include/linux/genhd.h && echo "#define HAVE_HD_STRUCT_IN_FLIGHT"
# Since 2.6, queue_lock was a pointer. Since 5.0 it's a non-pointer member
grep -q "spinlock_t[[:space:]]*queue_lock" $KDIR/include/linux/blkdev.h && echo "#define HAVE_RQ_Q_NO_PTR"
# nr_congestion_on was removed
# Perhaps needs more research to figure out if there's a replacement.
grep -q "unsigned int.*nr_congestion_on;" $KDIR/include/linux/blkdev.h && echo "#define HAVE_RQ_NR_CONGESTION"
# 5.6
grep -qw 'struct proc_ops' $KDIR/include/linux/proc_fs.h && echo "#define HAVE_PROC_OPS"
# 5.7
# blk_alloc_queue changes from (gfp_t) -> (make_request_fn make_request, int node_id
grep -qw 'blk_alloc_queue.make_request_fn' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BLK_ALLOC_QUEUE_RQFN"
grep -qw INCLUDE_VERMAGIC $KDIR/include/linux/vermagic.h && echo "#define HAVE_INCLUDE_VERMAGIC"
# blk_mq_make_request (GPL) was initially internal only (v3.13), and exported as GPL only in 5.8, and renamed in 5.9
# blk_mq_make_request was renamed to blk_mq_submit_bio in v5.9 (still GPL), but export removed in
# v5.9.2, when device mapper nolonger needed it.
grep -qw 'blk_qc_t blk_mq_make_request' $KDIR/include/linux/blk-mq.h && echo "#define HAVE_BLK_MQ_MAKE_REQUEST"
# 5.9
# removes typedef for queue_rq_fn
grep -qw 'blk_status_t .\*queue_rq..struct blk_mq_hw_ctx' $KDIR/include/linux/blk-mq.h && echo "#define NO_QUEUE_RQ_FN"
# blk_alloc_queue changes from (make_request_fn make_request, int node_id) -> (int node_id
grep -qw 'blk_alloc_queue.int node_id' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BLK_ALLOC_QUEUE_NODE_ID"
# obsoletes q->make_request_fn in favor of block_device_operations->submit_bio
grep -qw 'blk_qc_t ..submit_bio. .struct bio .bio.;' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BDOPS_SUBMIT_BIO_QC_T"
grep -qw 'mm_segment_t force_uaccess_begin' $KDIR/include/linux/uaccess.h && echo "#define HAVE_FORCE_UACCESS_BEGIN"
grep -qw 'blkdev_get_by_dev' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BLKDEV_GET_BY_DEV"
# 5.10
grep -qw 'bdev_whole' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BDEV_WHOLE"
# 5.11
# removes bd_contains, replaces it with bdev_whole macro
grep -qw 'define bdev_whole' $KDIR/include/linux/blk_types.h && echo "#define HAVE_BDEV_WHOLE"
grep -qw 'int freeze_bdev' $KDIR/include/linux/blkdev.h && echo "#define HAVE_INT_FREEZE_BDEV"
grep -qw bdev_nr_sectors $KDIR/include/linux/genhd.h && echo "#define HAVE_BDEV_NR_SECTORS"
# 5.12
# Explicit set budget after a get
grep -qw 'set_rq_budget_token' $KDIR/include/linux/blk-mq.h && echo "#define HAVE_MQ_OPS_SET_RQ_BUDGET_TOKEN"
grep -qw 'BIO_REMAPPED' $KDIR/include/linux/blk_types.h && echo "#define HAVE_BIO_REMAPPED"
# 5.16
# submit_bio drops the return value
grep -qw 'void ..submit_bio. .struct bio .bio.;' $KDIR/include/linux/blkdev.h && echo "#define HAVE_BDOPS_SUBMIT_BIO_VOID"

for symbol in kfree blk_mq_submit_bio; do
	addr=$($sudo awk "/ $symbol$/"'{print $1}' $SYSTEM_MAP)
	if [ -n "$addr" ]; then
		echo "#define HAVE_${symbol^^}"
		echo "#define ${symbol^^}_ADDR 0x${addr:-0}"
	fi
done

echo "#if defined HAVE_BLK_MQ_MAKE_REQUEST || defined HAVE_BLK_MQ_SUBMIT_BIO"
echo "#define WITH_MQ_SUPPORT"
echo "#endif"

exit 0
