/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include "filter.h"
#include "portability.h"
#include "table.h"

static struct block_device_filter _stcvsm_filter_table[];

#define STCVSM_REQFN(I) __stcvsm_reqfn_ ## I

#ifdef HAVE_BDOPS_SUBMIT_BIO
#define STCVSM_DEFINE_REQFN(I) \
	static blk_qc_t STCVSM_REQFN(I) (struct bio *b) \
	{ \
		return block_device_filter_reqfn(bio_request_queue(b), b, &_stcvsm_filter_table[I]); \
	}
#elif defined MAKE_REQUEST_RETVAL
#define STCVSM_DEFINE_REQFN(I) \
	static MAKE_REQUEST_RETTYPE STCVSM_REQFN(I) (struct request_queue *q, struct bio *b) \
	{ \
		return block_device_filter_reqfn(q, b, &_stcvsm_filter_table[I]); \
	}
#else
#define STCVSM_DEFINE_REQFN(I) \
	static void STCVSM_REQFN(I) (struct request_queue *q, struct bio *b) \
	{ \
		block_device_filter_reqfn(q, b, &_stcvsm_filter_table[I]); \
	}
#endif

#define STCVSM_TABLE_ENTRY(I) { \
	.bdf_reqfn = STCVSM_REQFN(I), \
	.idx = I }

#define STCVSM_DEFINE_REQFN_SET(I)	\
	STCVSM_DEFINE_REQFN(0x##I##0)	\
	STCVSM_DEFINE_REQFN(0x##I##1)	\
	STCVSM_DEFINE_REQFN(0x##I##2)	\
	STCVSM_DEFINE_REQFN(0x##I##3)	\
	STCVSM_DEFINE_REQFN(0x##I##4)	\
	STCVSM_DEFINE_REQFN(0x##I##5)	\
	STCVSM_DEFINE_REQFN(0x##I##6)	\
	STCVSM_DEFINE_REQFN(0x##I##7)	\
	STCVSM_DEFINE_REQFN(0x##I##8)	\
	STCVSM_DEFINE_REQFN(0x##I##9)	\
	STCVSM_DEFINE_REQFN(0x##I##A)	\
	STCVSM_DEFINE_REQFN(0x##I##B)	\
	STCVSM_DEFINE_REQFN(0x##I##C)	\
	STCVSM_DEFINE_REQFN(0x##I##D)	\
	STCVSM_DEFINE_REQFN(0x##I##E)	\
	STCVSM_DEFINE_REQFN(0x##I##F)	\

STCVSM_DEFINE_REQFN_SET(0)
STCVSM_DEFINE_REQFN_SET(1)
STCVSM_DEFINE_REQFN_SET(2)
STCVSM_DEFINE_REQFN_SET(3)
STCVSM_DEFINE_REQFN_SET(4)
STCVSM_DEFINE_REQFN_SET(5)
STCVSM_DEFINE_REQFN_SET(6)
STCVSM_DEFINE_REQFN_SET(7)
STCVSM_DEFINE_REQFN_SET(8)
STCVSM_DEFINE_REQFN_SET(9)
STCVSM_DEFINE_REQFN_SET(A)
STCVSM_DEFINE_REQFN_SET(B)
STCVSM_DEFINE_REQFN_SET(C)
STCVSM_DEFINE_REQFN_SET(D)
STCVSM_DEFINE_REQFN_SET(E)
STCVSM_DEFINE_REQFN_SET(F)

#define STCVSM_TABLE_ENTRY_SET(I)	\
	STCVSM_TABLE_ENTRY(0x##I##0),	\
	STCVSM_TABLE_ENTRY(0x##I##1),	\
	STCVSM_TABLE_ENTRY(0x##I##2),	\
	STCVSM_TABLE_ENTRY(0x##I##3),	\
	STCVSM_TABLE_ENTRY(0x##I##4),	\
	STCVSM_TABLE_ENTRY(0x##I##5),	\
	STCVSM_TABLE_ENTRY(0x##I##6),	\
	STCVSM_TABLE_ENTRY(0x##I##7),	\
	STCVSM_TABLE_ENTRY(0x##I##8),	\
	STCVSM_TABLE_ENTRY(0x##I##9),	\
	STCVSM_TABLE_ENTRY(0x##I##A),	\
	STCVSM_TABLE_ENTRY(0x##I##B),	\
	STCVSM_TABLE_ENTRY(0x##I##C),	\
	STCVSM_TABLE_ENTRY(0x##I##D),	\
	STCVSM_TABLE_ENTRY(0x##I##E),	\
	STCVSM_TABLE_ENTRY(0x##I##F)	\


static struct block_device_filter _stcvsm_filter_table[] = {
	STCVSM_TABLE_ENTRY_SET(0),
	STCVSM_TABLE_ENTRY_SET(1),
	STCVSM_TABLE_ENTRY_SET(2),
	STCVSM_TABLE_ENTRY_SET(3),
	STCVSM_TABLE_ENTRY_SET(4),
	STCVSM_TABLE_ENTRY_SET(5),
	STCVSM_TABLE_ENTRY_SET(6),
	STCVSM_TABLE_ENTRY_SET(7),
	STCVSM_TABLE_ENTRY_SET(8),
	STCVSM_TABLE_ENTRY_SET(9),
	STCVSM_TABLE_ENTRY_SET(A),
	STCVSM_TABLE_ENTRY_SET(B),
	STCVSM_TABLE_ENTRY_SET(C),
	STCVSM_TABLE_ENTRY_SET(D),
	STCVSM_TABLE_ENTRY_SET(E),
	STCVSM_TABLE_ENTRY_SET(F)
};

struct block_device_filter *stcvsm_filter_table(void)
{
	return &_stcvsm_filter_table[0];
}
