/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCVSM_VDIFF_H
#define STCVSM_VDIFF_H

#ifdef USE_RANGE_TABLE
#include "stcvsm_os.h"
#include "range_table.h"
#endif

struct _VSNAP_IOCTL_REANIMATE_BUFF;

#define VDIFF_SHUTDOWN		0
#define VDIFF_RESET		1
struct vdiff_context {
	/* Refcount */
	atomic_t vd_refs;

	/* Original Device info */
	dev_t vd_dev;
	struct block_device *vd_bdev;	/* Optional */
	sector_t vd_sector_start;	/* Partition offset */
	sector_t vd_sectors;		/* Partition size */

	/* Tracking data */
	u8 VDiffGuid[16];
	u8 VDiffFileGuid[16];
	u32 VDiffResetCount;
	unsigned long vd_flags;

	/* Bitmap */
#ifdef USE_RANGE_TABLE
	struct range_page_table *vd_range_table;
	struct lookaside_list *vd_lookaside;
#else
	unsigned long *vd_bitmap;
#endif

	/* Pointer back to bdfilter */
	struct block_device_filter *bdfilter;

	/* If !NULL, points to original device vdiff */
	struct vdiff_context *vdiff_orig;
};

/** Allocates an empty vdiff structure, sized to match the bdev.  Opens the bdev to prevent
 * repartitioning.
 * @return Returns a vdiff_context, or an ERR_PTR.
 */
struct vdiff_context *alloc_vdiff(struct block_device *bdev);
void rean_vdiff(struct vdiff_context *vdiff, u8 *chain_guid, u8 *file_guid, u32 reset_count);
void free_vdiff(struct vdiff_context *vdiff);
struct vdiff_context *get_vdiff_bio(struct block_device_filter *bdfilter, struct bio *bio);
struct vdiff_context *get_vdiff_bdev(struct block_device_filter *bdfilter,
		struct block_device *bdev);
void rollback_vdiff(struct vdiff_context *vdiff);
struct vdiff_context *copy_reset_vdiff(struct block_device_filter *bdfilter,
		struct vdiff_context *vdiff_old, dev_t snap_dev);
void put_vdiff(struct vdiff_context *vdiff);
void shutdown_vdiff(struct vdiff_context *vdiff);
void set_vdiff_range(struct vdiff_context *vdiff, sector_t start_sect, sector_t nr_sects);
sector_t count_vdiff(struct vdiff_context *vdiff);
void get_vdiff_range(struct vdiff_context *vdiff, sector_t start_sect, sector_t *nr_sects,
		bool *diffed);
int stcvsm_ioctl_vdiff(dev_t dev, unsigned int cmd, void *param);
int _stcvsm_ioctl_vdiff(struct block_device_filter *bdfilter, struct block_device *bdev,
		unsigned int cmd, void *param);

#endif
